"""TotalConnectClient() in this file is the primary class of this package.
Instantiate it like this:

usercodes = { 'default': '1234' }
client = TCC.TotalConnectClient(username, password, usercodes)

for location in client.locations:
    ### do stuff with this location
"""

import logging
import time

import zeep
import zeep.cache
import zeep.transports
import requests.exceptions

from .const import ArmType, _ResultCode
from .exceptions import (
    AuthenticationError,
    BadResultCodeError,
    InvalidSessionError,
    RetryableTotalConnectError,
    TotalConnectError,
)
from .location import TotalConnectLocation
from .user import TotalConnectUser

DEFAULT_USERCODE = "-1"

LOGGER = logging.getLogger(__name__)


class TotalConnectClient:
    """Client for Total Connect."""

    TIMEOUT = 60  # seconds until SOAP I/O will fail

    def __init__(         # pylint: disable=too-many-arguments
        self,
        username,
        password,
        usercodes=None,
        auto_bypass_battery=False,
        retry_delay=6,  # seconds between retries
    ):
        """Initialize."""
        self.times = {}
        self.time_start = time.time()
        self.soap_client = None

        self.username = username
        self.password = password
        self.usercodes = usercodes or {}
        self.auto_bypass_low_battery = auto_bypass_battery
        self.retry_delay = retry_delay

        self.token = None
        self._invalid_credentials = False
        self._module_flags = None
        self._user = None
        self._locations = {}
        self._locations_unfetched = {}

        self.authenticate()

        self.times["__init__"] = time.time() - self.time_start

    @property
    def locations(self):
        """Raises an exception if the panel cannot be reached to retrieve
        metadata or details. This can be retried later and will succeed
        if/when the panel becomes reachable.
        """
        # to_fetch is needed because items() is invalidated by del
        to_fetch = list(self._locations_unfetched.items())
        for (locationid, location) in to_fetch:
            try:
                location.get_partition_details()
                location.get_zone_details()
                location.get_panel_meta_data()
                # if we get here, it has been fetched successfully
                del self._locations_unfetched[locationid]
            except Exception:
                LOGGER.error(f"exception during initial fetch of {locationid}")
                raise
        assert not self._locations_unfetched
        return self._locations

    def __str__(self):
        """Return a text string that is printable."""
        data = (
            f"CLIENT\n\n"
            f"Username: {self.username}\n"
            f"Password: {'[hidden]' if self.password else '[unset]'}\n"
            f"Usercode: {self.usercodes}\n"
            f"Auto Bypass Low Battery: {self.auto_bypass_low_battery}\n"
            f"Invalid Credentials: {self._invalid_credentials}\n"
            f"Module Flags:\n"
        )

        for key, value in self._module_flags.items():
            data = data + f"  {key}: {value}\n"

        data = data + str(self._user)

        locations = f"LOCATIONS: {len(self.locations)}\n\n"
        for location in self.locations:
            locations = locations + str(self.locations[location])

        return data + locations

    def times_as_string(self):
        """Return a string with times."""
        self.times["total running time"] = time.time() - self.time_start
        msg = "total-connect-client time info (seconds):\n"
        for key, value in self.times.items():
            msg = msg + f"  {key}: {value}\n"

        return msg

    def _raise_for_retry(self, response):
        """Used internally to determine which responses should be retried in
        request().
        """
        rc = _ResultCode.from_response(response)
        if rc == _ResultCode.INVALID_SESSION:
            raise InvalidSessionError("invalid session", response)
        if rc == _ResultCode.INVALID_SESSIONID:
            raise InvalidSessionError("invalid session ID", response)
        if rc == _ResultCode.CONNECTION_ERROR:
            raise RetryableTotalConnectError("connection error", response)
        if rc == _ResultCode.FAILED_TO_CONNECT:
            raise RetryableTotalConnectError("failed to connect with panel", response)

    def raise_for_resultcode(self, response):
        """If response.ResultCode indicates success, return and do nothing.
        If it indicates an authentication error, raise AuthenticationError.
        """
        rc = _ResultCode.from_response(response)
        if rc in (
                _ResultCode.SUCCESS,
                _ResultCode.ARM_SUCCESS,
                _ResultCode.DISARM_SUCCESS,
                _ResultCode.SESSION_INITIATED,
        ):
            return
        self._raise_for_retry(response)
        if rc in (
                _ResultCode.BAD_USER_OR_PASSWORD,
                _ResultCode.AUTHENTICATION_FAILED,
                _ResultCode.USER_CODE_UNAVAILABLE,
        ):
            raise AuthenticationError(rc.name, response)
        raise BadResultCodeError(rc.name, response)

    def _send_one_request(self, operation_name, args):
        LOGGER.debug(f"sending API request {operation_name}{args}")
        operation_proxy = self.soap_client.service[operation_name]
        return zeep.helpers.serialize_object(operation_proxy(*args))

    API_ENDPOINT = "https://rs.alarmnet.com/TC21api/tc2.asmx?WSDL"
    API_APP_ID = "14588"
    API_APP_VERSION = "1.0.34"

    def request(self, operation_name, args, attempts_remaining=5):
        """Send a SOAP request. args is a list or tuple defining the
        parameters to the operation.
        """
        is_first_request = attempts_remaining == 5
        attempts_remaining -= 1
        if not self.soap_client:
            transport = zeep.transports.Transport(
                cache=zeep.cache.InMemoryCache(timeout=3600),
                timeout=self.TIMEOUT,  # for loading WSDL and xsd documents
                operation_timeout=self.TIMEOUT,  # for operations (POST/GET)
            )
            self.soap_client = zeep.Client(self.API_ENDPOINT, transport=transport)
        try:
            response = self._send_one_request(operation_name, args)
            self._raise_for_retry(response)
            return response
        # To retry an exception that could be raised during the
        # request, add it to an except block here, depending on what
        # you want to have happen. The first block just retries and
        # logs. The second block causes reauthentication.
        except (RetryableTotalConnectError, requests.exceptions.RequestException) as err:
            if attempts_remaining <= 0:
                raise
            if isinstance(err, RetryableTotalConnectError):
                msg = f"{self.username} {operation_name}{args} {err.args[0]} on response"
            else:
                msg = f"{self.username} {operation_name}{args} {err} on request"
            if is_first_request:
                LOGGER.info(f"{msg}: {attempts_remaining} retries remaining")
            else:
                LOGGER.debug(f"{msg}: {attempts_remaining} retries remaining")
            time.sleep(self.retry_delay)
        except InvalidSessionError:
            if attempts_remaining <= 0:
                raise
            LOGGER.info(f"reauthenticating {self.username}: {attempts_remaining} retries remaining")
            self.token = None
            self.authenticate()
        return self.request(operation_name, args, attempts_remaining)

    def authenticate(self):
        """Login to the system. Upon success, self.token is a valid credential
        for further API calls, and self._user and self.locations are valid.
        self.locations will not be refreshed if it was non-empty on entry.
        """
        start_time = time.time()
        if self._invalid_credentials:
            raise AuthenticationError(
                f"not authenticating: password already failed for user {self.username}"
            )

        # LoginAndGetSessionDetails is very slow, so only use it when necessary
        operation_name = (
            "AuthenticateUserLogin" if self._locations else "LoginAndGetSessionDetails"
        )
        response = self.request(operation_name, (
            self.username, self.password, self.API_APP_ID, self.API_APP_VERSION
        ))
        try:
            self.raise_for_resultcode(response)
        except AuthenticationError:
            self._invalid_credentials = True
            self.token = None
            raise

        self.token = response["SessionID"]
        if not self._locations:
            self._module_flags = dict(
                x.split("=") for x in response["ModuleFlags"].split(",")
            )
            self._user = TotalConnectUser(response["UserInfo"])
            self._locations_unfetched = self._make_locations(response)
            self._locations = self._locations_unfetched.copy()
            if not self._locations:
                raise TotalConnectError("no locations found", response)
        LOGGER.info(f"{self.username} authenticated: {len(self._locations)} locations")
        self.times["authenticate()"] = time.time() - start_time

    def validate_usercode(self, device_id, usercode):
        """Return True if the usercode is valid for the device."""
        response = self.request("ValidateUserCode", (self.token, device_id, str(usercode)))
        if _ResultCode.from_response(response) in (
            _ResultCode.USER_CODE_INVALID,
            _ResultCode.USER_CODE_UNAVAILABLE,
        ):
            LOGGER.warning(f"usercode {usercode} invalid for device {device_id}")
            return False
        self.raise_for_resultcode(response)
        return True

    def is_logged_in(self):
        """Return true if the client is logged into the Total Connect service
        with valid credentials.
        """
        return self.token is not None

    def log_out(self):
        """Upon return, we are logged out. Raises TotalConnectError if we
        still might be logged in.
        """
        if self.is_logged_in():
            response = self.request("Logout", (self.token,))
            self.raise_for_resultcode(response)
            LOGGER.info("Logout Successful")
            self.token = None

    def get_number_locations(self):
        """Return the number of locations.  Home Assistant needs a way
        to force the locations to load inside a callable function.
        """
        return len(self.locations)

    def _make_locations(self, response):
        """Return a dict mapping LocationID to TotalConnectLocation."""
        start_time = time.time()
        new_locations = {}

        for locationinfo in (response["Locations"] or {}).get("LocationInfoBasic", {}):
            location_id = locationinfo["LocationID"]
            location = TotalConnectLocation(locationinfo, self)
            new_locations[location_id] = location

            location.auto_bypass_low_battery = self.auto_bypass_low_battery

            # set the usercode for the location
            usercode = (
                self.usercodes.get(location_id) or  # noqa: W504
                self.usercodes.get(str(location_id)) or self.usercodes.get("default")
            )
            if usercode:
                location.usercode = usercode
            else:
                LOGGER.warning(f"no usercode for location {location_id}")
                location.usercode = DEFAULT_USERCODE

        self.times["_make_locations()"] = time.time() - start_time
        return new_locations


class ArmingHelper:
    """
    For a partition or location, you can call its arm() or disarm() method directly.
       Example: partition.arm(ArmType.AWAY)

    Alternatively, you can use ArmingHelper.
       Example: ArmingHelper(partition).arm_away()
    """

    def __init__(self, partition_or_location):
        self.armable = partition_or_location

    def arm_away(self):
        """Arm the system (Away)."""
        self.armable.arm(ArmType.AWAY)

    def arm_stay(self):
        """Arm the system (Stay)."""
        self.armable.arm(ArmType.STAY)

    def arm_stay_instant(self):
        """Arm the system (Stay - Instant)."""
        self.armable.arm(ArmType.STAY_INSTANT)

    def arm_away_instant(self):
        """Arm the system (Away - Instant)."""
        self.armable.arm(ArmType.AWAY_INSTANT)

    def arm_stay_night(self):
        """Arm the system (Stay - Night)."""
        self.armable.arm(ArmType.STAY_NIGHT)

    def disarm(self):
        self.armable.disarm()



"""Exceptions raised by TotalConnectClient."""


class TotalConnectError(Exception):
    pass


class AuthenticationError(TotalConnectError):
    pass


class InvalidSessionError(AuthenticationError):
    pass


class BadResultCodeError(TotalConnectError):
    pass


class RetryableTotalConnectError(TotalConnectError):
    """These errors are likely to resolve themselves if the action is retried.
    If an error requires some other action (such as reauthenticating an
    expired session) before being retried, it is not "retryable."
    """


class PartialResponseError(RetryableTotalConnectError):
    """Raised if the response is missing a section that it is always supposed to have.
    Because the TotalConnect servers are flaky, these are rather frequent.
    """



"""Total Connect Device."""


class TotalConnectDevice:
    """Device class for Total Connect."""

    def __init__(self, info):
        """Initialize device based on DeviceInfoBasic."""
        self.deviceid = info.get("DeviceID")
        self.name = info.get("DeviceName")
        self.class_id = info.get("DeviceClassID")
        self.serial_number = info.get("DeviceSerialNumber")
        self.security_panel_type_id = info.get("SecurityPanelTypeID")
        self.serial_text = info.get("DeviceSerialText")

        flags = info.get("DeviceFlags")
        if flags is None:
            self.flags = {}
        else:
            self.flags = dict(x.split("=") for x in flags.split(","))

    def __str__(self):
        """Return a string that is printable."""
        data = (
            f"DEVICE {self.deviceid} - {self.name}\n"
            f"  ClassID: {self.class_id}\n"
            f"  Security Panel Type ID: {self.security_panel_type_id}\n"
            f"  Serial Number: {self.serial_number}\n"
            f"  Serial Text: {self.serial_text}\n"
            f"  Device Flags:\n"
        )

        for key, value in self.flags.items():
            data = data + f"    {key}: {value}\n"

        return data



from distutils.core import setup
from setuptools import setup, find_packages

setup(
    name="total_connect_client",
    version="2021.11.1",
    description="Interact with Total Connect 2 alarm systems",
    author="Craig J. Midwinter",
    author_email="craig.j.midwinter@gmail.com",
    url="https://github.com/craigjmidwinter/total-connect-client",
    download_url="https://github.com/craigjmidwinter/total-connect-client",
    keywords=["alarm", "TotalConnect"],
    package_data={"": ["data/*.json"]},
    install_requires=["zeep>=4.1.0"],
    python_requires=">=3.7",
    packages=["total_connect_client"],
    include_package_data=True,  # use MANIFEST.in during install
    zip_safe=False,
)



"""Total Connect Zone."""

import logging
from enum import Enum, IntFlag

from .const import PROJECT_URL
from .exceptions import TotalConnectError

LOGGER = logging.getLogger(__name__)


class ZoneStatus(IntFlag):
    NORMAL      =   0
    BYPASSED    =   1
    FAULT       =   2
    TROUBLE     =   8  # is also Tampered
    LOW_BATTERY =  64
    TRIGGERED   = 256


class ZoneType(Enum):
    """
    These are the "standard" Honeywell zone types. However some panels such
    as the Lynx 7000 report most security zones as zone type 3.

    https://www.alarmliquidators.com/content/Vista%2021IP-%20Programming%20Guide.pdf
    http://techresource.online/training/ssnw/honeywell/zone-types
    """

    SECURITY          = 0  # for Vista, zone type 0 is not used
    ENTRY_EXIT1       = 1  # starts countdown timer #1
    ENTRY_EXIT2       = 2  # like ENTRY_EXIT1 but uses timer #2
    PERIMETER         = 3  # zone type 3 usually triggers an immediate alarm...
    PROA7_SECURITY    = 3  # but some panels like Lynx 7000 report timed zones as 3
    INTERIOR_FOLLOWER = 4  # inactive when armed STAY
    TROUBLE_ALARM     = 5  # trouble by day, alarm by night
    SILENT_24HR       = 6  # 24-hour silent alarm (often used for police/hold-up)
    AUDIBLE_24HR      = 7  # 24-hour audible alarm (often used for police)
    AUX_24HR          = 8  # no local siren but keypad beeps (often used for medical)
    FIRE_SMOKE        = 9
    INTERIOR_DELAY    = 10  # inactive when armed STAY, otherwise like ENTRY_EXIT1
    MONITOR           = 12  # e.g. temperature or flood
    CARBON_MONOXIDE   = 14
    PROA7_MEDICAL     = 15
    FIRE_W_VERIFICATION = 16  # must trigger twice to cause an alarm
    LYRIC_LOCAL_ALARM = 89

    # According to the VISTA docs, these can be programmed via downloader software
    # or from a keypad using data fields *182-*185

    VISTA_CONFIGURABLE_90 = 90
    VISTA_CONFIGURABLE_91 = 91
    VISTA_CONFIGURABLE_92 = 92
    VISTA_CONFIGURABLE_93 = 93


class TotalConnectZone:
    """Do not create instances of this class yourself."""

    def __init__(self, zone):
        """Initialize."""
        self.zoneid = zone.get("ZoneID")
        self.partition = None
        self.status = None
        self.zone_type_id = None
        self.can_be_bypassed = None
        self.battery_level = None
        self.signal_strength = None
        self.sensor_serial_number = None
        self.loop_number = None
        self.response_type = None
        self.alarm_report_state = None
        self.supervision_type = None
        self.chime_state = None
        self.device_type = None
        self._update(zone)

    def __str__(self):
        """Return a string that is printable."""
        return (
            f"Zone {self.zoneid} - {self.description}\n"
            f"  Partition: {self.partition}\t\t"
            f"Zone Type: {self.zone_type_id}\t"
            f"CanBeBypassed: {self.can_be_bypassed}\t"
            f"Status: {self.status}\n"
            f"  Battery Level: {self.battery_level}\t"
            f"Signal Stength: {self.signal_strength}\n"
            f"  Serial Number: {self.sensor_serial_number}\t"
            f"Loop: {self.loop_number}\t"
            f"Response Type: {self.response_type}\n"
            f"  Supervision Type: {self.supervision_type}\t"
            f"Alarm Report State: {self.alarm_report_state}\n"
            f"  Chime State: {self.chime_state}\t"
            f"Device Type: {self.device_type}\n\n"
        )

    def is_bypassed(self):
        """Return true if the zone is bypassed."""
        return self.status & ZoneStatus.BYPASSED > 0

    def is_faulted(self):
        """Return true if the zone is faulted."""
        return self.status & ZoneStatus.FAULT > 0

    def is_tampered(self):
        """Return true if zone is tampered."""
        return self.status & ZoneStatus.TROUBLE > 0

    def is_low_battery(self):
        """Return true if low battery."""
        return self.status & ZoneStatus.LOW_BATTERY > 0

    def is_troubled(self):
        """Return true if zone is troubled."""
        return self.status & ZoneStatus.TROUBLE > 0

    def is_triggered(self):
        """Return true if zone is triggered."""
        return self.status & ZoneStatus.TRIGGERED > 0

    def is_type_button(self):
        """Return true if zone is a button."""

        # as seen so far, any security zone that cannot be bypassed is a button on a panel
        if self.is_type_security() and not self.can_be_bypassed:
            return True

        if self.zone_type_id in (
                ZoneType.PROA7_MEDICAL,
                ZoneType.AUDIBLE_24HR,
                ZoneType.SILENT_24HR,
        ):
            return True

        return False

    def is_type_security(self):
        """Return true if zone type is security."""

        return self.zone_type_id in (
            ZoneType.SECURITY,
            ZoneType.ENTRY_EXIT1,
            ZoneType.ENTRY_EXIT2,
            ZoneType.PERIMETER,
            ZoneType.INTERIOR_FOLLOWER,
            ZoneType.TROUBLE_ALARM,
            ZoneType.SILENT_24HR,
            ZoneType.AUDIBLE_24HR,
            ZoneType.INTERIOR_DELAY,
            ZoneType.LYRIC_LOCAL_ALARM,
        )

    def is_type_motion(self):
        """Return true if zone type is motion."""
        return self.zone_type_id == ZoneType.INTERIOR_FOLLOWER

    def is_type_fire(self):
        """Return true if zone type is fire or smoke."""
        # TODO: why is ZoneType.MONITOR here? it's not a fire or smoke zone type
        return self.zone_type_id in (ZoneType.FIRE_SMOKE, ZoneType.MONITOR)

    def is_type_carbon_monoxide(self):
        """Return true if zone type is carbon monoxide."""
        return self.zone_type_id == ZoneType.CARBON_MONOXIDE

    def is_type_medical(self):
        """Return true if zone type is medical."""
        return self.zone_type_id == ZoneType.PROA7_MEDICAL

    def _update(self, zone):
        """Update the zone."""
        assert zone
        zid = zone.get("ZoneID")
        assert self.zoneid == zid, (self.zoneid, zid)

        self.description = zone.get("ZoneDescription")
        # ZoneInfo gives 'PartitionID' but ZoneStatusInfoWithPartitionId gives 'PartitionId'
        if "PartitionId" in zone:
            # ...and PartitionId gives an int instead of a string
            self.partition = str(zone["PartitionId"])
        else:
            self.partition = zone.get("PartitionID")
        try:
            self.status = ZoneStatus(zone.get("ZoneStatus"))
        except ValueError:
            LOGGER.error(f"unknown ZoneStatus in {zone}: please report at {PROJECT_URL}/issues")
            raise TotalConnectError(f"unknown ZoneStatus in {zone}") from None
        self.can_be_bypassed = zone.get("CanBeBypassed")

        try:
            zid = zone.get("ZoneTypeId", self.zone_type_id)
            # TODO: if zid is None should we raise PartialResponseError?
            self.zone_type_id = None if zid is None else ZoneType(zid)
        except ValueError:
            LOGGER.error(f"unknown ZoneType {zid} in {zone}: please report at {PROJECT_URL}/issues")
            # if we get an unknown ZoneType we do not raise an exception, because
            # we know there are more zone types than we have in our enum, and
            # having an unknown ZoneType doesn't keep us from doing our work
            self.zone_type_id = zid

        self.battery_level = zone.get("Batterylevel", self.battery_level)
        self.signal_strength = zone.get("Signalstrength", self.signal_strength)
        info = zone.get("zoneAdditionalInfo")
        if info:
            self.sensor_serial_number = info.get("SensorSerialNumber")
            self.loop_number = info.get("LoopNumber")
            self.response_type = info.get("ResponseType")
            self.alarm_report_state = info.get("AlarmReportState")
            self.supervision_type = info.get("ZoneSupervisionType")
            self.chime_state = info.get("ChimeState")
            self.device_type = info.get("DeviceType")

    def _mark_as_bypassed(self):
        """Set is_bypassed status."""
        # TODO: when does this get reset to no longer bypassed?
        self.status |= ZoneStatus.BYPASSED



"""Total Connect Client constants."""

from enum import Enum

from .exceptions import BadResultCodeError


class ArmType(Enum):
    AWAY = 0
    STAY = 1
    STAY_INSTANT = 2
    AWAY_INSTANT = 3
    STAY_NIGHT = 4


class ArmingState(Enum):
    DISARMED = 10200
    DISARMED_BYPASS = 10211
    DISARMED_ZONE_FAULTED = 10214  # seems to apply to location, not to partition.  See issue #144

    ARMED_AWAY = 10201
    ARMED_AWAY_BYPASS = 10202
    ARMED_STAY = 10203
    ARMED_STAY_OTHER = 10226
    ARMED_STAY_BYPASS = 10204
    ARMED_AWAY_INSTANT = 10205
    ARMED_AWAY_INSTANT_BYPASS = 10206
    ARMED_STAY_INSTANT = 10209
    ARMED_STAY_INSTANT_BYPASS = 10210
    ARMED_STAY_NIGHT = 10218
    ARMED_CUSTOM_BYPASS = 10223

    ALARMING = 10207
    ALARMING_FIRE_SMOKE = 10212
    ALARMING_CARBON_MONOXIDE = 10213

    ARMING = 10307
    DISARMING = 10308

    def is_arming(self):
        """Return true if the system is in the process of arming."""
        return self == ArmingState.ARMING

    def is_disarming(self):
        """Return true if the system is in the process of disarming."""
        return self == ArmingState.DISARMING

    def is_pending(self):
        """Return true if the system is pending an action."""
        return self.is_disarming() or self.is_arming()

    def is_disarmed(self):
        """Return True if the system is disarmed."""
        return self in (
            ArmingState.DISARMED, ArmingState.DISARMED_BYPASS, ArmingState.DISARMED_ZONE_FAULTED
        )

    def is_armed_away(self):
        """Return True if the system is armed away in any way."""
        return self in (
            ArmingState.ARMED_AWAY,
            ArmingState.ARMED_AWAY_BYPASS,
            ArmingState.ARMED_AWAY_INSTANT,
            ArmingState.ARMED_AWAY_INSTANT_BYPASS,
        )

    def is_armed_custom_bypass(self):
        """Return True if the system is armed custom bypass in any way."""
        return self == ArmingState.ARMED_CUSTOM_BYPASS

    def is_armed_home(self):
        """Return True if the system is armed home/stay in any way."""
        return self in (
            ArmingState.ARMED_STAY,
            ArmingState.ARMED_STAY_BYPASS,
            ArmingState.ARMED_STAY_INSTANT,
            ArmingState.ARMED_STAY_INSTANT_BYPASS,
            ArmingState.ARMED_STAY_NIGHT,
            ArmingState.ARMED_STAY_OTHER,
        )

    def is_armed_night(self):
        """Return True if the system is armed night in any way."""
        return self == ArmingState.ARMED_STAY_NIGHT

    def is_armed(self):
        """Return True if the system is armed in any way."""
        return (
            self.is_armed_away() or self.is_armed_home() or  # noqa: W504
            self.is_armed_night() or self.is_armed_custom_bypass()
        )

    def is_triggered_police(self):
        """Return True if the system is triggered for police or medical."""
        return self == ArmingState.ALARMING

    def is_triggered_fire(self):
        """Return True if the system is triggered for fire or smoke."""
        return self == ArmingState.ALARMING_FIRE_SMOKE

    def is_triggered_gas(self):
        """Return True if the system is triggered for carbon monoxide."""
        return self == ArmingState.ALARMING_CARBON_MONOXIDE

    def is_triggered(self):
        """Return True if the system is triggered in any way."""
        return (
            self.is_triggered_fire() or self.is_triggered_gas() or self.is_triggered_police()
        )


class _ResultCode(Enum):
    """As suggested by the leading underscore, this class is not used by
    callers of the API.
    """

    @staticmethod
    def from_response(response_dict):
        try:
            return _ResultCode(response_dict["ResultCode"])
        except ValueError:
            raise BadResultCodeError(
                f"unknown result code {response_dict['ResultCode']}", response_dict
            ) from None

    SUCCESS = 0
    ARM_SUCCESS = 4500
    DISARM_SUCCESS = 4500
    SESSION_INITIATED = 4500

    BAD_USER_OR_PASSWORD = -50004
    INVALID_SESSIONID = -30002
    COMMAND_FAILED = -4502
    USER_CODE_UNAVAILABLE = -4114
    USER_CODE_INVALID = -4106
    FAILED_TO_CONNECT = -4104
    FEATURE_NOT_SUPPORTED = -120
    INVALID_SESSION = -102
    AUTHENTICATION_FAILED = -100
    CONNECTION_ERROR = 4101


PROJECT_URL = "https://github.com/craigjmidwinter/total-connect-client"
